


class DummyCharDisplayDriver(object):
    """Class used in tests"""

    def __init__(self):
        pass
    @property
    def display_type(self):
        return "char"
    

    def home(self):
        """Move the cursor back to its home (first line and first column)."""
        print("dummy lcd: home")

    def clear(self):
        """Clear the LCD."""
        print("dummy lcd: clear")

    def set_cursor(self, col, row):
        """Move the cursor to an explicit column and row position."""
        print("dummy lcd: set_cursor", col, row)

    def enable_display(self, enable):
        """Enable or disable the display.  Set enable to True to enable."""
        print("dummy lcd: enable", enable)

    def show_cursor(self, show):
        """Show or hide the cursor.  Cursor is shown if show is True."""
        print("dummy lcd: show_cursor", show)

    def blink(self, blink):
        """Turn on or off cursor blinking.  Set blink to True to enable blinking."""
        print("dummy lcd: blink", blink)

    def move_left(self):
        """Move display left one position."""
        print("dummy lcd: move_left")

    def move_right(self):
        """Move display right one position."""
        print("dummy lcd: right")

    def set_left_to_right(self):
        """Set text direction left to right."""
        print("dummy lcd: set_left_right")

    def set_right_to_left(self):
        """Set text direction right to left."""
        print("dummy lcd: set_right_to_Left")

    def autoscroll(self, autoscroll):
        """Autoscroll will 'right justify' text from the cursor if set True,
        otherwise it will 'left justify' the text.
        """
        print("dummy lcd: autoscroll")

    def message(self, text):
        """Write text to display.  Note that text can include newlines."""
        print("dummy lcd: message", text)

    def set_backlight(self, backlight):
        """Enable or disable the backlight.  If PWM is not enabled (default), a
        non-zero backlight value will turn on the backlight and a zero value will
        turn it off.  If PWM is enabled, backlight can be any value from 0.0 to
        1.0, with 1.0 being full intensity backlight.
        """
        print("dummy lcd: backlight", backlight)

    def create_char(self, location, pattern):
        """Fill one of the first 8 CGRAM locations with custom characters.
        The location parameter should be between 0 and 7 and pattern should
        provide an array of 8 bytes containing the pattern. E.g. you can easyly
        design your custom character at http://www.quinapalus.com/hd44780udg.html
        To show your custom character use eg. lcd.message('\x01')
        """
        print("dummy lcd: create_char", location, pattern)

    
class DummyBitmapDisplayDriver(object):
    """Base class for SSD1306-based OLED displays.  Implementors should subclass
    and provide an implementation for the _initialize function.
    """

    def __init__(self, width, height):
        self.width = width
        self.height = height
        self._pages = height//8
        self._buffer = [0]*(width*self._pages)
        
    @property
    def display_type(self):
        return "bitmap"
 
    def reset(self):
        """Reset the display."""
        print("bitmap display: reset")

    def display(self):
        """Write display buffer to physical display."""
        print("bitmap display: display")

    def image(self, image):
        """Set buffer to value of Python Imaging Library image.  The image should
        be in 1 bit mode and a size equal to the display size.
        """
        if image.mode != '1':
            raise ValueError('Image must be in mode 1.')
        imwidth, imheight = image.size
        if imwidth != self.width or imheight != self.height:
            raise ValueError('Image must be same dimensions as display ({0}x{1}).' \
                .format(self.width, self.height))
        print("bitmap display: image")
        image.save("dummydisplay.bmp")

    def clear(self):
        """Clear contents of image buffer."""
        print("bitmap display: clear")

    def set_contrast(self, contrast):
        """Sets the contrast of the display.  Contrast should be a value between
        0 and 255."""
        print("bitmap display: set_contrast", contrast)

    def dim(self, dim):
        """Adjusts contrast to dim the display if dim is True, otherwise sets the
        contrast to normal brightness if dim is False.
        """
        print("bitmap display: dim", dim)


# The MIT License (MIT)
#
# Copyright (c) 2018 Carter Nelson for Adafruit Industries
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
`ads1x15`
====================================================

CircuitPython base class driver for ADS1015/1115 ADCs.

* Author(s): Carter Nelson
"""

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_ADS1x15.git"

from adafruit_bus_device.i2c_device import I2CDevice

# pylint: disable=bad-whitespace
_ADS1X15_DEFAULT_ADDRESS            = 0x48
_ADS1X15_POINTER_CONVERSION         = 0x00
_ADS1X15_POINTER_CONFIG             = 0x01
_ADS1X15_CONFIG_OS_SINGLE           = 0x8000
_ADS1X15_CONFIG_MUX_OFFSET          = 12
_ADS1X15_CONFIG_COMP_QUE_DISABLE    = 0x0003
_ADS1X15_CONFIG_GAIN = {
    2/3: 0x0000,
    1:   0x0200,
    2:   0x0400,
    4:   0x0600,
    8:   0x0800,
    16:  0x0A00
}
# pylint: enable=bad-whitespace

class _Mode:
    """An enum-like class representing possible ADC operating modes."""
    # See datasheet "Operating Modes" section
    # values here are masks for setting MODE bit in Config Register
    #pylint: disable=too-few-public-methods
    CONTINUOUS = 0x0000
    SINGLE = 0x0100


class _ADS1x15(object):
    """Base functionality for ADS1x15 analog to digital converters."""

    def __init__(self, i2c, gain=1, data_rate=None, mode=_Mode.SINGLE,
                 address=_ADS1X15_DEFAULT_ADDRESS):
        #pylint: disable=too-many-arguments
        self.buf = bytearray(3)
        self._data_rate = self._gain = self._mode = None
        self.gain = gain
        self.data_rate = self._data_rate_default() if data_rate is None else data_rate
        self.mode = mode
        self.i2c_device = I2CDevice(i2c, address)

    @property
    def data_rate(self):
        """The data rate for ADC conversion in samples per second."""
        return self._data_rate

    @data_rate.setter
    def data_rate(self, rate):
        possible_rates = self.rates
        if rate not in possible_rates:
            raise ValueError("Data rate must be one of: {}".format(possible_rates))
        self._data_rate = rate

    @property
    def rates(self):
        """Possible data rate settings."""
        raise NotImplementedError('Subclass must implement rates property.')

    @property
    def rate_config(self):
        """Rate configuration masks."""
        raise NotImplementedError('Subclass must implement rate_config property.')

    @property
    def gain(self):
        """The ADC gain."""
        return self._gain

    @gain.setter
    def gain(self, gain):
        possible_gains = self.gains
        if gain not in possible_gains:
            raise ValueError("Gain must be one of: {}".format(possible_gains))
        self._gain = gain

    @property
    def gains(self):
        """Possible gain settings."""
        g = list(_ADS1X15_CONFIG_GAIN.keys())
        g.sort()
        return g

    @property
    def mode(self):
        """The ADC conversion mode."""
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode != Mode.CONTINUOUS and mode != Mode.SINGLE:
            raise ValueError("Unsupported mode.")
        self._mode = mode

    def read(self, pin, is_differential=False):
        """I2C Interface for ADS1x15-based ADCs reads.

        params:
            :param pin: individual or differential pin.
            :param bool is_differential: single-ended or differential read.
        """
        pin = pin if is_differential else pin + 0x04
        return self._read(pin)

    def _data_rate_default(self):
        """Retrieve the default data rate for this ADC (in samples per second).
        Should be implemented by subclasses.
        """
        raise NotImplementedError('Subclasses must implement _data_rate_default!')

    def _conversion_value(self, raw_adc):
        """Subclasses should override this function that takes the 16 raw ADC
        values of a conversion result and returns a signed integer value.
        """
        raise NotImplementedError('Subclass must implement _conversion_value function!')

    def _read(self, pin):
        """Perform an ADC read. Returns the signed integer result of the read."""
        config = _ADS1X15_CONFIG_OS_SINGLE
        config |= (pin & 0x07) << _ADS1X15_CONFIG_MUX_OFFSET
        config |= _ADS1X15_CONFIG_GAIN[self.gain]
        config |= self.mode
        config |= self.rate_config[self.data_rate]
        config |= _ADS1X15_CONFIG_COMP_QUE_DISABLE
        self._write_register(_ADS1X15_POINTER_CONFIG, config)

        if self.mode == Mode.SINGLE:
            while not self._conversion_complete():
                pass

        return self.get_last_result()

    def _conversion_complete(self):
        """Return status of ADC conversion."""
        # OS is bit 15
        # OS = 0: Device is currently performing a conversion
        # OS = 1: Device is not currently performing a conversion
        return self._read_register(_ADS1X15_POINTER_CONFIG) & 0x8000

    def get_last_result(self):
        """Read the last conversion result when in continuous conversion mode.
        Will return a signed integer value.
        """
        return self._conversion_value(self._read_register(_ADS1X15_POINTER_CONVERSION))

    def _write_register(self, reg, value):
        """Write 16 bit value to register."""
        self.buf[0] = reg
        self.buf[1] = (value >> 8) & 0xFF
        self.buf[2] = value & 0xFF
        with self.i2c_device as i2c:
            i2c.write(self.buf)

    def _read_register(self, reg):
        """Read 16 bit register value."""
        self.buf[0] = reg
        with self.i2c_device as i2c:
            i2c.write(self.buf, end=1, stop=False)
            i2c.readinto(self.buf, end=2)
        return self.buf[0] << 8 | self.buf[1]


_ADS1115_CONFIG_DR = {
    8:    0x0000,
    16:   0x0020,
    32:   0x0040,
    64:   0x0060,
    128:  0x0080,
    250:  0x00A0,
    475:  0x00C0,
    860:  0x00E0
}

# Pins
P0 = 0
P1 = 1
P2 = 2
P3 = 3

class ADS1115DeviceDriver(_ADS1x15):
    """Class for the ADS1115 16 bit ADC."""

    @property
    def bits(self):
        """The ADC bit resolution."""
        return 16

    @property
    def rates(self):
        """Possible data rate settings."""
        r = list(_ADS1115_CONFIG_DR.keys())
        r.sort()
        return r

    @property
    def rate_config(self):
        """Rate configuration masks."""
        return _ADS1115_CONFIG_DR

    def _data_rate_default(self):
        return 128

    def _conversion_value(self, raw_adc):
        raw_adc = raw_adc.to_bytes(2, "big")
        value = struct.unpack(">h", raw_adc)[0]
        return value


# Data sample rates
_ADS1015_CONFIG_DR = {
    128:   0x0000,
    250:   0x0020,
    490:   0x0040,
    920:   0x0060,
    1600:  0x0080,
    2400:  0x00A0,
    3300:  0x00C0
}

# Pins
P0 = 0
P1 = 1
P2 = 2
P3 = 3

class ADS1015DeviceDriver(_ADS1x15):
    """Class for the ADS1015 12 bit ADC."""

    @property
    def bits(self):
        """The ADC bit resolution."""
        return 12

    @property
    def rates(self):
        """Possible data rate settings."""
        r = list(_ADS1015_CONFIG_DR.keys())
        r.sort()
        return r

    @property
    def rate_config(self):
        """Rate configuration masks."""
        return _ADS1015_CONFIG_DR

    def _data_rate_default(self):
        return 1600

    def _conversion_value(self, raw_adc):
        raw_adc = raw_adc.to_bytes(2, "big")
        value = struct.unpack(">h", raw_adc)[0]
        return value >> 4


from random import randint
from kervi.hal import SensorDeviceDriver, I2CSensorDeviceDriver

class DummySensorDeviceDriver(SensorDeviceDriver):
    def __init__(self, **kwargs):
        SensorDeviceDriver.__init__(self)

        self.value = 0
        self.delta = kwargs.get("delta", 4)
        self._min = kwargs.get("min", 0)
        self._max = kwargs.get("max", 100)
        self._interval = kwargs.get("interval", 1)
        self._unit = kwargs.get("unit", "c")
        self._type = kwargs.get("type", "temperature")

    def read_value(self):
        self.value += self.delta
        if self.value > self._max:
            self.delta *= -1
            self.value = self._max
        if self.value < self._min:
            self.delta *= -1
            self.value = self._min
        return self.value

    @property
    def type(self):
        return self._type

    @property
    def unit(self):
        return self._unit

    @property
    def max(self):
        return self._max

    @property
    def min(self):
        return self._min


class DummyMultiDimSensorDeviceDriver(SensorDeviceDriver):
    def __init__(self):
        SensorDeviceDriver.__init__(self)
        self.value = 0
        self.delta = 4

    def read_value(self):
        self.value += self.delta
        if self.value == 100:
            self.delta *= -1
        if self.value == 0:
            self.delta *= -1
        return [self.value, self.value + 1, self.value +2]

    @property
    def dimensions(self):
        return 3

    @property
    def dimension_labels(self):
        return ["heading", "pitch", "roll"]

    @property
    def type(self):
        return "position"

    @property
    def unit(self):
        return "degree"

    @property
    def max(self):
        return 100

    @property
    def min(self):
        return 0


class DummyColorSensorDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, **kwargs):
        I2CSensorDeviceDriver.__init__(self, 0x20, 0)

        self.value = 0
        self.delta = kwargs.get("delta", 4)
        self._min = kwargs.get("min", 0)
        self._max = kwargs.get("max", 100)
        self._interval = kwargs.get("interval", 1)
        self._unit = kwargs.get("unit", "rgb")
        self._type = kwargs.get("type", "color")

    def read_value(self):
        value = [randint(0,255), randint(0,255), randint(0,255)]
        return value

    @property
    def type(self):
        return self._type

    @property
    def unit(self):
        return self._unit

    @property
    def max(self):
        return self._max

    @property
    def min(self):
        return self._min

    @property
    def value_type(self):
        return "color"



# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT

import time
import random
from kervi.hal import DeviceChannelOutOfBoundsError, DACValueOutOfBoundsError
from kervi.hal import ChannelPollingThread
from kervi.hal.gpio import IGPIODeviceDriver, CHANNEL_TYPE_GPIO, CHANNEL_TYPE_ANALOG_IN, CHANNEL_TYPE_ANALOG_OUT
from kervi.core.utility.thread import KerviThread
from kervi import spine

class GPIODummyDriver(IGPIODeviceDriver):

    # Constructor
    def __init__(self, gpio_id = "dummy_gpio", manual = False):
        IGPIODeviceDriver.__init__(self, gpio_id)
        self._manual = manual
        self._listeners = []
        self._pin_map = {
            "AIN0": 0,
            "AIN1": 1,
            "AIN2": 2,
            "AIN3": 3,
            "AOUT0": 4,
            "AOUT1": 5,
            "AOUT2": 6,
            "AOUT3": 7
        }
        self._values = {
            0: 0,
            1: 0,
            2: 0,
            3: 0,
            4: 0,
            5: 0,
            6: 0,
            7: 0
        }

    def _get_channel_type(self, channel):
        if int(channel) in [0, 1, 2, 3]:
            return CHANNEL_TYPE_ANALOG_IN
        elif int(channel) in [4, 5, 6, 7]:
            return CHANNEL_TYPE_ANALOG_OUT

    def _get_channel_names(self):
        return self._pin_map.keys()

    def _map_pin(self, channel):
        if isinstance(channel, int):
            return channel
        elif str(channel) in self._pin_map.keys():
            return self._pin_map[str(channel)]
        else:
            return int(channel)
    
    @property
    def device_name(self):
        return "DummyGPIO"
 
    def get(self, channel):
        checked_channel = self._check_channel_adc(channel)
 
        if channel in [0, 1, 2, 3]:
            if self._manual:
                self._values[checked_channel] = self._values[4 + checked_channel] 
            else:
                old_reading = self._values[checked_channel]
                new_reading = random.randrange(10) / 100
                new_reading = (-1)**random.randrange(2) * new_reading
                self._values[checked_channel] = old_reading + new_reading

        return self._values[checked_channel]

    def set(self, channel, state):
        checked_channel = self._check_channel_dac(channel)
        checked_val = self._check_dac_val(channel, state)
        self._values[checked_channel] = checked_val 

    def define_as_input(self, channel, pullup=False):
        self._check_channel_adc(channel)

    def define_as_output(self, channel):
        self._check_channel_dac(channel)

    def listen(self, channel, callback, polling_time=.1):
        self._listeners += [ChannelPollingThread(channel, self, callback, polling_time)]

    # Check if ADC channel number is within bounds
    def _check_channel_adc(self, chan):
        if type(chan) is str:
            chan = int(chan)
        if type(chan) is not int:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan < 0:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan > 3:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        return chan

    def _check_channel_dac(self, chan):
        if type(chan) is str:
            chan = int(chan)
        if type(chan) is not int:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan < 4:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan > 7:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        return chan

    # Check if DAC output value is within bounds
    def _check_dac_val(self, channel, val):
        if type(val) is not float:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        elif val < 0:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        elif val > 1:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        return val



# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT

import time
from kervi.hal import I2CGPIODeviceDriver, DeviceChannelOutOfBoundsError, DACValueOutOfBoundsError
from kervi.hal import ChannelPollingThread
from kervi.hal.gpio import CHANNEL_TYPE_GPIO, CHANNEL_TYPE_ANALOG_IN, CHANNEL_TYPE_ANALOG_OUT
from kervi.core.utility.thread import KerviThread
from kervi import spine
AIN0 = 0
AIN1 = 1
AIN2 = 2
AIN3 = 3
AOUT = 4

class PCF8591Driver(I2CGPIODeviceDriver):

    # Constructor
    def __init__(self, address = 0x48, bus=0, gpio_id="PCF8591"):
        I2CGPIODeviceDriver.__init__(self, address, bus, gpio_id)
        self._dac_enabled = 0x00
        self._listeners = []
        self._pin_map = {
            "AIN0": 0,
            "AIN1": 1, 
            "AIN2": 2, 
            "AIN3": 3, 
            "AOUT": 4
        }

    def _get_channel_type(self, channel):
        if channel in [0, 1, 2, 3]:
            return CHANNEL_TYPE_ANALOG_IN
        elif channel == 4:
            return CHANNEL_TYPE_ANALOG_OUT

    def _get_channel_names(self):
        return self._pin_map.keys()

    def _map_pin(self, channel):
        if isinstance(channel, int):
            return channel
        elif str(channel) in self._pin_map.keys():
            return self._pin_map[str(channel)]
        else:
            return int(channel)
    
    @property
    def device_name(self):
        return "PFC8591"
 
    def get(self, channel):
        """Read single ADC Channel"""
        checked_channel = self._check_channel_no(channel)
        self.i2c.write_raw8(checked_channel  | self._dac_enabled)
        reading = self.i2c.read_raw8() 
        reading = self.i2c.read_raw8() 
        return reading / 255.0

    def set(self, channel, state):
        """Set DAC value and enable output"""
        checked_val = self._check_dac_val(channel, state)
        self._dac_enabled = 0x40
        self.i2c.write8(self._dac_enabled, checked_val * 255)

    def define_as_input(self, channel, pullup=False):
        self._check_channel_no(channel)

    def define_as_output(self, channel):
        if not channel == 0:
            raise DeviceChannelOutOfBoundsError(self.device_name, channel)

    def listen(self, channel, callback, polling_time=.1):
        self._listeners += [ChannelPollingThread(channel, self, callback, polling_time)]

    # Enable DAC output
    def enable_dac(self):
        self._dac_enabled = 0x40
        self.i2c.write_raw8(self._dac_enabled)

    # Disable DAC output
    def disable_dac(self):
        self.__dac_enabled = 0x00
        self.i2c.write_raw8(self.__dac_enabled)

    # Check if ADC channel number is within bounds
    def _check_channel_no(self, chan):
        if type(chan) is not int:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan < 0:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        elif chan > 3:
            raise DeviceChannelOutOfBoundsError(self.device_name, chan)
        return chan

    # Check if DAC output value is within bounds
    def _check_dac_val(self, channel, val):
        if type(val) is not float:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        elif val < 0:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        elif val > 1:
            raise DACValueOutOfBoundsError(self.device_name, channel, val)
        return val



from kervi.hal import I2CSensorDeviceDriver

VL6180X_DEFAULT_I2C_ADDR = 0x29

VL6180X_REG_IDENTIFICATION_MODEL_ID = 0x000
VL6180X_REG_SYSTEM_INTERRUPT_CONFIG = 0x014
VL6180X_REG_SYSTEM_INTERRUPT_CLEAR = 0x015
VL6180X_REG_SYSTEM_FRESH_OUT_OF_RESET = 0x016
VL6180X_REG_SYSRANGE_START = 0x018
VL6180X_REG_SYSALS_START = 0x038
VL6180X_REG_SYSALS_ANALOGUE_GAIN = 0x03F
VL6180X_REG_SYSALS_INTEGRATION_PERIOD_HI = 0x040
VL6180X_REG_SYSALS_INTEGRATION_PERIOD_LO = 0x041
VL6180X_REG_RESULT_ALS_VAL = 0x050
VL6180X_REG_RESULT_RANGE_VAL = 0x062
VL6180X_REG_RESULT_RANGE_STATUS = 0x04d
VL6180X_REG_RESULT_INTERRUPT_STATUS_GPIO = 0x04f


VL6180X_ALS_GAIN_1 = 0x06
VL6180X_ALS_GAIN_1_25 = 0x05
VL6180X_ALS_GAIN_1_67 = 0x04
VL6180X_ALS_GAIN_2_5 = 0x03
VL6180X_ALS_GAIN_5 = 0x02
VL6180X_ALS_GAIN_10 = 0x01
VL6180X_ALS_GAIN_20 = 0x00
VL6180X_ALS_GAIN_40 = 0x07

VL6180X_ERROR_NONE = 0
VL6180X_ERROR_SYSERR_1 = 1
VL6180X_ERROR_SYSERR_5 = 5
VL6180X_ERROR_ECEFAIL = 6
VL6180X_ERROR_NOCONVERGE = 7
VL6180X_ERROR_RANGEIGNORE = 8
VL6180X_ERROR_SNR = 11
VL6180X_ERROR_RAWUFLOW = 12
VL6180X_ERROR_RAWOFLOW = 13
VL6180X_ERROR_RANGEUFLOW = 14
VL6180X_ERROR_RANGEOFLOW = 15


class _VL6180X(I2CSensorDeviceDriver):
    def __init__(self, type, address, bus):
        I2CSensorDeviceDriver.__init__(self, address, bus)
        self._type = type
        self._gain = VL6180X_ALS_GAIN_1
        
    @property
    def type(self):
        if self._type == "distance":
            return "distance"
        else:
            return "lux"

    @property
    def unit(self):
        if self._type == "distance":
            return "mm"
        else:
            return "lx"

    @property
    def max(self):
        if self._type == "distance":
            return 200
        else:
            return 1000000
        

    @property
    def min(self):
        return 0

    def read_value(self):
        if self._type == "distance":
            return self.read_range()
        else:
            return self.read_lux()

    def setup(self):
        if self.i2c.read_U8(VL6180X_REG_IDENTIFICATION_MODEL_ID) != 0xB4:
            return False

        self.load_settings()

        self.i2c.write8(VL6180X_REG_SYSTEM_FRESH_OUT_OF_RESET, 0x00)

        return True

    def load_settings(self):
        # private settings from page 24 of app note
        self.i2c.write8( 0x0207, 0x01)
        self.i2c.write8( 0x0208, 0x01)
        self.i2c.write8( 0x0097, 0xfd)
        self.i2c.write8( 0x0096, 0x00)
        self.i2c.write8( 0x00e3, 0x00)
        self.i2c.write8( 0x00e4, 0x04)
        self.i2c.write8( 0x00e5, 0x02)
        self.i2c.write8( 0x00e6, 0x01)
        self.i2c.write8( 0x00e7, 0x03)
        self.i2c.write8( 0x00f5, 0x02)
        self.i2c.write8( 0x00d9, 0x05)
        self.i2c.write8( 0x00db, 0xce)
        self.i2c.write8( 0x00dc, 0x03)
        self.i2c.write8( 0x00dd, 0xf8)
        self.i2c.write8( 0x009f, 0x00)
        self.i2c.write8( 0x00a3, 0x3c)
        self.i2c.write8( 0x00b7, 0x00)
        self.i2c.write8( 0x00bb, 0x3c)
        self.i2c.write8( 0x00b2, 0x09)
        self.i2c.write8( 0x00ca, 0x09)
        self.i2c.write8( 0x0198, 0x01)
        self.i2c.write8( 0x01b0, 0x17)
        self.i2c.write8( 0x01ad, 0x00)
        self.i2c.write8( 0x00ff, 0x05)
        self.i2c.write8( 0x0100, 0x05)
        self.i2c.write8( 0x0199, 0x05)
        self.i2c.write8( 0x01a6, 0x1b)
        self.i2c.write8( 0x01ac, 0x3e)
        self.i2c.write8( 0x01a7, 0x1f)
        self.i2c.write8( 0x0030, 0x00)

        # Recommended : Public registers - See data sheet for more detail
        self.i2c.write8( 0x0011, 0x10)  # Enables polling for 'New Sample ready' when measurement
        # completes
        self.i2c.write8( 0x010a, 0x30)  # Set the averaging sample period (compromise between
        # lower noise and increased execution time)
        self.i2c.write8( 0x003f, 0x46)  # Sets the light and dark gain (upper nibble). Dark gain
        # should not be changed.
        self.i2c.write8( 0x0031, 0xFF)  # sets the # of range measurements after which auto
        # calibration of system is performed
        self.i2c.write8( 0x0040, 0x63)  # Set ALS integration time to 100ms
        self.i2c.write8( 0x002e, 0x01)  # perform a single temperature calibration of the ranging
        # sensor

        # Optional: Public registers - See data sheet for more detail
        self.i2c.write8( 0x001b, 0x09)  # Set default ranging inter-measurement period to 100ms
        self.i2c.write8( 0x003e, 0x31)  # Set default ALS inter-measurement period to 500ms
        self.i2c.write8( 0x0014, 0x24)  # Configures interrupt on 'New Sample Ready threshold event'


    def read_range(self):
        # wait for device to be ready for range measurement
        while not (self.i2c.read_U8(VL6180X_REG_RESULT_RANGE_STATUS) & 0x01): pass

        # Start a range measurement
        self.i2c.write8(VL6180X_REG_SYSRANGE_START, 0x01)

        # Poll until bit 2 is set
        while not (self.i2c.read_U8(VL6180X_REG_RESULT_INTERRUPT_STATUS_GPIO) & 0x04): pass

        # read range in mm
        range = self.i2c.read_U16(VL6180X_REG_RESULT_RANGE_VAL)

        # clear interrupt
        self.i2c.write8(VL6180X_REG_SYSTEM_INTERRUPT_CLEAR, 0x07)

        return range

    #def read_range_status(self):
    #    return self.b.read_byte_data(self._addr, VL6180X_REG_RESULT_RANGE_STATUS) >> 4


    def read_lux(self):
        gain = self._gain
        reg = self.i2c.read_U8(VL6180X_REG_SYSTEM_INTERRUPT_CONFIG)
        reg &= ~0x38
        reg |= (0x4 << 3)  # IRQ on ALS ready
        self.i2c.write8( VL6180X_REG_SYSTEM_INTERRUPT_CONFIG, reg)

        # 100 ms integration period
        self.i2c.write8( VL6180X_REG_SYSALS_INTEGRATION_PERIOD_HI, 0)
        self.i2c.write8( VL6180X_REG_SYSALS_INTEGRATION_PERIOD_LO, 100)

        # analog gain
        if gain > VL6180X_ALS_GAIN_40:
            gain = VL6180X_ALS_GAIN_40

        self.i2c.write8( VL6180X_REG_SYSALS_ANALOGUE_GAIN, 0x40 | gain)

        # start ALS
        self.i2c.write8( VL6180X_REG_SYSALS_START, 0x1)

        # Poll until "New Sample Ready threshold event" is set
        while 4 != ((self.i2c.read_U8(VL6180X_REG_RESULT_INTERRUPT_STATUS_GPIO) >> 3) & 0x7): pass

        # read lux!
        lux = self.i2c.read_U16(VL6180X_REG_RESULT_ALS_VAL)

        # clear interrupt
        self.i2c.write8( VL6180X_REG_SYSTEM_INTERRUPT_CLEAR, 0x07)

        lux *= 0.32  # calibrated count/lux
        if gain == VL6180X_ALS_GAIN_1:
            pass
        elif gain == VL6180X_ALS_GAIN_1_25:
            lux /= 1.25
        elif gain == VL6180X_ALS_GAIN_1_67:
            lux /= 1.76
        elif gain == VL6180X_ALS_GAIN_2_5:
            lux /= 2.5
        elif gain == VL6180X_ALS_GAIN_5:
            lux /= 5
        elif gain == VL6180X_ALS_GAIN_10:
            lux /= 10
        elif gain == VL6180X_ALS_GAIN_20:
            lux /= 20
        elif gain == VL6180X_ALS_GAIN_40:
            lux /= 20

        lux *= 100
        lux /= 100  # integration time in ms

        return lux

class VL6180XDistanceDeviceDriver(_VL6180X):
    def __init__(self, address = VL6180X_DEFAULT_I2C_ADDR, bus=0):
            _VL6180X.__init__(self, "distance", address, bus)

class VL6180XLuxDeviceDriver(_VL6180X):
    def __init__(self, gain = VL6180X_ALS_GAIN_1, address = VL6180X_DEFAULT_I2C_ADDR, bus=0):
            _VL6180X.__init__(self, "lux", address, bus)




# The MIT License (MIT)
#
# Copyright (c) 2017 Tony DiCola for Adafruit Industries
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#This driver is derived from Tony DoCola's work and adapted to the api of the kervi framework.

import time
try:
    import struct
except ImportError:
    import ustruct as struct
from kervi.hal import get_i2c, SensorDeviceDriver

# Internal constants and register values:
# pylint: disable=bad-whitespace
_LSM9DS1_ADDRESS_ACCELGYRO       = 0x6B
_LSM9DS1_ADDRESS_MAG             = 0x1E
_LSM9DS1_XG_ID                   = 0b01101000
_LSM9DS1_MAG_ID                  = 0b00111101
_LSM9DS1_ACCEL_MG_LSB_2G         = 0.061
_LSM9DS1_ACCEL_MG_LSB_4G         = 0.122
_LSM9DS1_ACCEL_MG_LSB_8G         = 0.244
_LSM9DS1_ACCEL_MG_LSB_16G        = 0.732
_LSM9DS1_MAG_MGAUSS_4GAUSS       = 0.14
_LSM9DS1_MAG_MGAUSS_8GAUSS       = 0.29
_LSM9DS1_MAG_MGAUSS_12GAUSS      = 0.43
_LSM9DS1_MAG_MGAUSS_16GAUSS      = 0.58
_LSM9DS1_GYRO_DPS_DIGIT_245DPS   = 0.00875
_LSM9DS1_GYRO_DPS_DIGIT_500DPS   = 0.01750
_LSM9DS1_GYRO_DPS_DIGIT_2000DPS  = 0.07000
_LSM9DS1_TEMP_LSB_DEGREE_CELSIUS = 8 # 1°C = 8, 25° = 200, etc.
_LSM9DS1_REGISTER_WHO_AM_I_XG    = 0x0F
_LSM9DS1_REGISTER_CTRL_REG1_G    = 0x10
_LSM9DS1_REGISTER_CTRL_REG2_G    = 0x11
_LSM9DS1_REGISTER_CTRL_REG3_G    = 0x12
_LSM9DS1_REGISTER_TEMP_OUT_L     = 0x15
_LSM9DS1_REGISTER_TEMP_OUT_H     = 0x16
_LSM9DS1_REGISTER_STATUS_REG     = 0x17
_LSM9DS1_REGISTER_OUT_X_L_G      = 0x18
_LSM9DS1_REGISTER_OUT_X_H_G      = 0x19
_LSM9DS1_REGISTER_OUT_Y_L_G      = 0x1A
_LSM9DS1_REGISTER_OUT_Y_H_G      = 0x1B
_LSM9DS1_REGISTER_OUT_Z_L_G      = 0x1C
_LSM9DS1_REGISTER_OUT_Z_H_G      = 0x1D
_LSM9DS1_REGISTER_CTRL_REG4      = 0x1E
_LSM9DS1_REGISTER_CTRL_REG5_XL   = 0x1F
_LSM9DS1_REGISTER_CTRL_REG6_XL   = 0x20
_LSM9DS1_REGISTER_CTRL_REG7_XL   = 0x21
_LSM9DS1_REGISTER_CTRL_REG8      = 0x22
_LSM9DS1_REGISTER_CTRL_REG9      = 0x23
_LSM9DS1_REGISTER_CTRL_REG10     = 0x24
_LSM9DS1_REGISTER_OUT_X_L_XL     = 0x28
_LSM9DS1_REGISTER_OUT_X_H_XL     = 0x29
_LSM9DS1_REGISTER_OUT_Y_L_XL     = 0x2A
_LSM9DS1_REGISTER_OUT_Y_H_XL     = 0x2B
_LSM9DS1_REGISTER_OUT_Z_L_XL     = 0x2C
_LSM9DS1_REGISTER_OUT_Z_H_XL     = 0x2D
_LSM9DS1_REGISTER_WHO_AM_I_M     = 0x0F
_LSM9DS1_REGISTER_CTRL_REG1_M    = 0x20
_LSM9DS1_REGISTER_CTRL_REG2_M    = 0x21
_LSM9DS1_REGISTER_CTRL_REG3_M    = 0x22
_LSM9DS1_REGISTER_CTRL_REG4_M    = 0x23
_LSM9DS1_REGISTER_CTRL_REG5_M    = 0x24
_LSM9DS1_REGISTER_STATUS_REG_M   = 0x27
_LSM9DS1_REGISTER_OUT_X_L_M      = 0x28
_LSM9DS1_REGISTER_OUT_X_H_M      = 0x29
_LSM9DS1_REGISTER_OUT_Y_L_M      = 0x2A
_LSM9DS1_REGISTER_OUT_Y_H_M      = 0x2B
_LSM9DS1_REGISTER_OUT_Z_L_M      = 0x2C
_LSM9DS1_REGISTER_OUT_Z_H_M      = 0x2D
_LSM9DS1_REGISTER_CFG_M          = 0x30
_LSM9DS1_REGISTER_INT_SRC_M      = 0x31
_MAGTYPE                         = True
_XGTYPE                          = False
_SENSORS_GRAVITY_STANDARD        = 9.80665

# User facing constants/module globals.
ACCELRANGE_2G                = (0b00 << 3)
ACCELRANGE_16G               = (0b01 << 3)
ACCELRANGE_4G                = (0b10 << 3)
ACCELRANGE_8G                = (0b11 << 3)
MAGGAIN_4GAUSS               = (0b00 << 5)  # +/- 4 gauss
MAGGAIN_8GAUSS               = (0b01 << 5)  # +/- 8 gauss
MAGGAIN_12GAUSS              = (0b10 << 5)  # +/- 12 gauss
MAGGAIN_16GAUSS              = (0b11 << 5)  # +/- 16 gauss
GYROSCALE_245DPS             = (0b00 << 3)  # +/- 245 degrees/s rotation
GYROSCALE_500DPS             = (0b01 << 3)  # +/- 500 degrees/s rotation
GYROSCALE_2000DPS            = (0b11 << 3)  # +/- 2000 degrees/s rotation
# pylint: enable=bad-whitespace


def _twos_comp(val, bits):
    # Convert an unsigned integer in 2's compliment form of the specified bit
    # length to its signed integer value and return it.
    if val & (1 << (bits - 1)) != 0:
        return val - (1 << bits)
    return val


class _LSM9DS1():
    """Driver for the LSM9DS1 accelerometer, magnetometer, gyroscope."""

    def __init__(self):
        self._BUFFER = bytearray(6)
        # soft reset & reboot accel/gyro
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG8, 0x05)
        # soft reset & reboot magnetometer
        self._write_u8(_MAGTYPE, _LSM9DS1_REGISTER_CTRL_REG2_M, 0x0C)
        time.sleep(0.01)
        # Check ID registers.
        if self._read_u8(_XGTYPE, _LSM9DS1_REGISTER_WHO_AM_I_XG) != _LSM9DS1_XG_ID or \
           self._read_u8(_MAGTYPE, _LSM9DS1_REGISTER_WHO_AM_I_M) != _LSM9DS1_MAG_ID:
            raise RuntimeError('Could not find LSM9DS1, check wiring!')
        # enable gyro continuous
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG1_G, 0xC0) # on XYZ
        # Enable the accelerometer continous
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG5_XL, 0x38)
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG6_XL, 0xC0)
        # enable mag continuous
        self._write_u8(_MAGTYPE, _LSM9DS1_REGISTER_CTRL_REG3_M, 0x00)
        # Set default ranges for the various sensors
        self._accel_mg_lsb = None
        self._mag_mgauss_lsb = None
        self._gyro_dps_digit = None
        self.accel_range = ACCELRANGE_2G
        self.mag_gain = MAGGAIN_4GAUSS
        self.gyro_scale = GYROSCALE_245DPS

    @property
    def accel_range(self):
        """The accelerometer range.  Must be a value of:
          - ACCELRANGE_2G
          - ACCELRANGE_4G
          - ACCELRANGE_8G
          - ACCELRANGE_16G
        """
        reg = self._read_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG6_XL)
        return (reg & 0b00011000) & 0xFF

    @accel_range.setter
    def accel_range(self, val):
        assert val in (ACCELRANGE_2G, ACCELRANGE_4G, ACCELRANGE_8G,
                       ACCELRANGE_16G)
        reg = self._read_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG6_XL)
        reg = (reg & ~(0b00011000)) & 0xFF
        reg |= val
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG6_XL, reg)
        if val == ACCELRANGE_2G:
            self._accel_mg_lsb = _LSM9DS1_ACCEL_MG_LSB_2G
        elif val == ACCELRANGE_4G:
            self._accel_mg_lsb = _LSM9DS1_ACCEL_MG_LSB_4G
        elif val == ACCELRANGE_8G:
            self._accel_mg_lsb = _LSM9DS1_ACCEL_MG_LSB_8G
        elif val == ACCELRANGE_16G:
            self._accel_mg_lsb = _LSM9DS1_ACCEL_MG_LSB_16G

    @property
    def mag_gain(self):
        """The magnetometer gain.  Must be a value of:
          - MAGGAIN_4GAUSS
          - MAGGAIN_8GAUSS
          - MAGGAIN_12GAUSS
          - MAGGAIN_16GAUSS
        """
        reg = self._read_u8(_MAGTYPE, _LSM9DS1_REGISTER_CTRL_REG2_M)
        return (reg & 0b01100000) & 0xFF

    @mag_gain.setter
    def mag_gain(self, val):
        assert val in (MAGGAIN_4GAUSS, MAGGAIN_8GAUSS, MAGGAIN_12GAUSS,
                       MAGGAIN_16GAUSS)
        reg = self._read_u8(_MAGTYPE, _LSM9DS1_REGISTER_CTRL_REG2_M)
        reg = (reg & ~(0b01100000)) & 0xFF
        reg |= val
        self._write_u8(_MAGTYPE, _LSM9DS1_REGISTER_CTRL_REG2_M, reg)
        if val == MAGGAIN_4GAUSS:
            self._mag_mgauss_lsb = _LSM9DS1_MAG_MGAUSS_4GAUSS
        elif val == MAGGAIN_8GAUSS:
            self._mag_mgauss_lsb = _LSM9DS1_MAG_MGAUSS_8GAUSS
        elif val == MAGGAIN_12GAUSS:
            self._mag_mgauss_lsb = _LSM9DS1_MAG_MGAUSS_12GAUSS
        elif val == MAGGAIN_16GAUSS:
            self._mag_mgauss_lsb = _LSM9DS1_MAG_MGAUSS_16GAUSS

    @property
    def gyro_scale(self):
        """The gyroscope scale.  Must be a value of:
          - GYROSCALE_245DPS
          - GYROSCALE_500DPS
          - GYROSCALE_2000DPS
        """
        reg = self._read_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG1_G)
        return (reg & 0b00011000) & 0xFF

    @gyro_scale.setter
    def gyro_scale(self, val):
        assert val in (GYROSCALE_245DPS, GYROSCALE_500DPS, GYROSCALE_2000DPS)
        reg = self._read_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG1_G)
        reg = (reg & ~(0b00011000)) & 0xFF
        reg |= val
        self._write_u8(_XGTYPE, _LSM9DS1_REGISTER_CTRL_REG1_G, reg)
        if val == GYROSCALE_245DPS:
            self._gyro_dps_digit = _LSM9DS1_GYRO_DPS_DIGIT_245DPS
        elif val == GYROSCALE_500DPS:
            self._gyro_dps_digit = _LSM9DS1_GYRO_DPS_DIGIT_500DPS
        elif val == GYROSCALE_2000DPS:
            self._gyro_dps_digit = _LSM9DS1_GYRO_DPS_DIGIT_2000DPS

    def read_accel_raw(self):
        """Read the raw accelerometer sensor values and return it as a
        3-tuple of X, Y, Z axis values that are 16-bit unsigned values.  If you
        want the acceleration in nice units you probably want to use the
        accelerometer property!
        """
        # Read the accelerometer
        self._read_bytes(_XGTYPE, 0x80 | _LSM9DS1_REGISTER_OUT_X_L_XL, 6,  self._BUFFER)
        raw_x, raw_y, raw_z = struct.unpack_from('<hhh', self._BUFFER[0:6])
        return (raw_x, raw_y, raw_z)

    @property
    def acceleration(self):
        """The accelerometer X, Y, Z axis values as a 3-tuple of
        m/s^2 values.
        """
        raw = self.read_accel_raw()
        return map(lambda x: x * self._accel_mg_lsb / 1000.0 * _SENSORS_GRAVITY_STANDARD,
                   raw)

    def read_mag_raw(self):
        """Read the raw magnetometer sensor values and return it as a
        3-tuple of X, Y, Z axis values that are 16-bit unsigned values.  If you
        want the magnetometer in nice units you probably want to use the
        magnetometer property!
        """
        # Read the magnetometer
        self._read_bytes(_MAGTYPE, 0x80 | _LSM9DS1_REGISTER_OUT_X_L_M, 6, self._BUFFER)
        raw_x, raw_y, raw_z = struct.unpack_from('<hhh', self._BUFFER[0:6])
        return (raw_x, raw_y, raw_z)

    @property
    def magnetic(self):
        """The magnetometer X, Y, Z axis values as a 3-tuple of
        gauss values.
        """
        raw = self.read_mag_raw()
        return map(lambda x: x * self._mag_mgauss_lsb / 1000.0, raw)

    def read_gyro_raw(self):
        """Read the raw gyroscope sensor values and return it as a
        3-tuple of X, Y, Z axis values that are 16-bit unsigned values.  If you
        want the gyroscope in nice units you probably want to use the
        gyroscope property!
        """
        # Read the gyroscope
        self._read_bytes(_XGTYPE, 0x80 | _LSM9DS1_REGISTER_OUT_X_L_G, 6,  self._BUFFER)
        raw_x, raw_y, raw_z = struct.unpack_from('<hhh', self._BUFFER[0:6])
        return (raw_x, raw_y, raw_z)

    @property
    def gyro(self):
        """The gyroscope X, Y, Z axis values as a 3-tuple of
        degrees/second values.
        """
        raw = self.read_gyro_raw()
        return map(lambda x: x * self._gyro_dps_digit, raw)

    def read_temp_raw(self):
        """Read the raw temperature sensor value and return it as a 12-bit
        signed value.  If you want the temperature in nice units you probably
        want to use the temperature property!
        """
        # Read temp sensor
        self._read_bytes(_XGTYPE, 0x80 | _LSM9DS1_REGISTER_TEMP_OUT_L, 2, self._BUFFER)
        temp = ((self._BUFFER[1] << 8) | self._BUFFER[0]) >> 4
        return _twos_comp(temp, 12)

    @property
    def temperature(self):
        """The temperature of the sensor in degrees Celsius."""
        # This is just a guess since the starting point (21C here) isn't documented :(
        # See discussion from:
        #  https://github.com/kriswiner/LSM9DS1/issues/3
        temp = self.read_temp_raw()
        temp = 27.5 + temp/16
        return temp

    def _read_u8(self, sensor_type, address):
        # Read an 8-bit unsigned value from the specified 8-bit address.
        # The sensor_type boolean should be _MAGTYPE when talking to the
        # magnetometer, or _XGTYPE when talking to the accel or gyro.
        # MUST be implemented by subclasses!
        raise NotImplementedError()

    def _read_bytes(self, sensor_type, address, count, buf):
        # Read a count number of bytes into buffer from the provided 8-bit
        # register address.  The sensor_type boolean should be _MAGTYPE when
        # talking to the magnetometer, or _XGTYPE when talking to the accel or
        # gyro.  MUST be implemented by subclasses!
        raise NotImplementedError()

    def _write_u8(self, sensor_type, address, val):
        # Write an 8-bit unsigned value to the specified 8-bit address.
        # The sensor_type boolean should be _MAGTYPE when talking to the
        # magnetometer, or _XGTYPE when talking to the accel or gyro.
        # MUST be implemented by subclasses!
        raise NotImplementedError()


class _LSM9DS1_I2C(_LSM9DS1):
    """Driver for the LSM9DS1 connect over I2C."""
    def __init__(self, acclgyro_address=_LSM9DS1_ADDRESS_ACCELGYRO, mag_address=_LSM9DS1_ADDRESS_MAG, bus=None):
        self._mag_device = get_i2c(mag_address, bus)
        self._xg_device = get_i2c(acclgyro_address, bus)
        super().__init__()

    def _read_u8(self, sensor_type, address):
        if sensor_type == _MAGTYPE:
            device = self._mag_device
        else:
            device = self._xg_device
        return device.read_U8(address)

    def _read_bytes(self, sensor_type, address, count, buf):
        if sensor_type == _MAGTYPE:
            device = self._mag_device
        else:
            device = self._xg_device
        r= device.read_list(address, count)
        buf[:] = r

    def _write_u8(self, sensor_type, address, val):
        if sensor_type == _MAGTYPE:
            device = self._mag_device
        else:
            device = self._xg_device
        device.write8(address, val)

class LSM9DS1AccelerationDeviceDriver(SensorDeviceDriver):
    def __init__(self, accel_range=ACCELRANGE_2G, acclgyro_address=_LSM9DS1_ADDRESS_ACCELGYRO, mag_address=_LSM9DS1_ADDRESS_MAG, bus=None):
        SensorDeviceDriver.__init__(self)
        self._device = _LSM9DS1_I2C(acclgyro_address,mag_address, bus)
        self._device.accel_range = accel_range

    @property
    def dimensions(self):
        return 3

    @property
    def value_type(self):
        return "number"
    

    @property
    def dimension_labels(self):
        return ["x","y", "z"]
    
    @property
    def type(self):
        return "acceleration"

    @property
    def unit(self):
        return "m/s^2"

    def read_value(self):
        x,y,z = self._device.acceleration
        return [x, y, z]

class LSM9DS1GyroDeviceDriver(SensorDeviceDriver):
    def __init__(self, gyro_scale=GYROSCALE_245DPS, acclgyro_address=_LSM9DS1_ADDRESS_ACCELGYRO, mag_address=_LSM9DS1_ADDRESS_MAG, bus=None):
        SensorDeviceDriver.__init__(self)
        self._device = _LSM9DS1_I2C(acclgyro_address,mag_address, bus)
        self._device.gyro_scale = gyro_scale

    @property
    def value_type(self):
        return "number"

    @property
    def dimensions(self):
        return 3

    @property
    def dimension_labels(self):
        return ["x","y", "z"]
    
    @property
    def type(self):
        return "gyro"

    @property
    def unit(self):
        return "degrees/second"

    def read_value(self):
        x,y,z = self._device.gyro
        return [x, y, z]

        
class LSM9DS1MagneticDeviceDriver(SensorDeviceDriver):
    def __init__(self, gain=MAGGAIN_4GAUSS, acclgyro_address=_LSM9DS1_ADDRESS_ACCELGYRO, mag_address=_LSM9DS1_ADDRESS_MAG, bus=None):
        SensorDeviceDriver.__init__(self)
        self._device = _LSM9DS1_I2C(acclgyro_address,mag_address, bus)
        self._device.mag_gain = gain

    @property
    def dimensions(self):
        return 3

    @property
    def dimension_labels(self):
        return ["x","y", "z"]
    
    @property
    def type(self):
        return "magnetic"

    @property
    def unit(self):
        return "gauss"

    @property
    def value_type(self):
        return "number"

    def read_value(self):
        x,y,z = self._device.magnetic
        return [x, y, z]



# Copyright (c) 2014 Adafruit Industries
# Author: Tony DiCola
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import division
import logging
import time
from kervi import hal


# Constants
SSD1306_I2C_ADDRESS = 0x3C    # 011110+SA0+RW - 0x3C or 0x3D
SSD1306_SETCONTRAST = 0x81
SSD1306_DISPLAYALLON_RESUME = 0xA4
SSD1306_DISPLAYALLON = 0xA5
SSD1306_NORMALDISPLAY = 0xA6
SSD1306_INVERTDISPLAY = 0xA7
SSD1306_DISPLAYOFF = 0xAE
SSD1306_DISPLAYON = 0xAF
SSD1306_SETDISPLAYOFFSET = 0xD3
SSD1306_SETCOMPINS = 0xDA
SSD1306_SETVCOMDETECT = 0xDB
SSD1306_SETDISPLAYCLOCKDIV = 0xD5
SSD1306_SETPRECHARGE = 0xD9
SSD1306_SETMULTIPLEX = 0xA8
SSD1306_SETLOWCOLUMN = 0x00
SSD1306_SETHIGHCOLUMN = 0x10
SSD1306_SETSTARTLINE = 0x40
SSD1306_MEMORYMODE = 0x20
SSD1306_COLUMNADDR = 0x21
SSD1306_PAGEADDR = 0x22
SSD1306_COMSCANINC = 0xC0
SSD1306_COMSCANDEC = 0xC8
SSD1306_SEGREMAP = 0xA0
SSD1306_CHARGEPUMP = 0x8D
SSD1306_EXTERNALVCC = 0x1
SSD1306_SWITCHCAPVCC = 0x2

# Scrolling constants
SSD1306_ACTIVATE_SCROLL = 0x2F
SSD1306_DEACTIVATE_SCROLL = 0x2E
SSD1306_SET_VERTICAL_SCROLL_AREA = 0xA3
SSD1306_RIGHT_HORIZONTAL_SCROLL = 0x26
SSD1306_LEFT_HORIZONTAL_SCROLL = 0x27
SSD1306_VERTICAL_AND_RIGHT_HORIZONTAL_SCROLL = 0x29
SSD1306_VERTICAL_AND_LEFT_HORIZONTAL_SCROLL = 0x2A


class _SSD1306I2CBase(object):
    """Base class for SSD1306-based OLED displays.  Implementors should subclass
    and provide an implementation for the _initialize function.
    """

    def __init__(self, width, height, rst_gpio=None, i2c_bus=None, address=SSD1306_I2C_ADDRESS):
        self._i2c = hal.get_i2c(address, i2c_bus)
        self.width = width
        self.height = height
        self._pages = height//8
        self._buffer = [0]*(width*self._pages)
        # Default to platform GPIO if not provided.
        self._rst = rst_gpio
        if self._rst:
            self._rst.define_as_output()

        self.begin()

    @property
    def display_type(self):
        return "bitmap"
    
    def _initialize(self):
        raise NotImplementedError

    def command(self, c):
        """Send command byte to display."""
        control = 0x00   # Co = 0, DC = 0
        self._i2c.write8(control, c)

    def data(self, c):
        """Send byte of data to display."""
        # I2C write.
        control = 0x40   # Co = 0, DC = 0
        self._i2c.write8(control, c)

    def begin(self, vccstate=SSD1306_SWITCHCAPVCC):
        """Initialize display."""
        # Save vcc state.
        self._vccstate = vccstate
        # Reset and initialize display.
        self.reset()
        self._initialize()
        # Turn on the display.
        self.command(SSD1306_DISPLAYON)

    def reset(self):
        """Reset the display."""
        if self._rst is None:
            return
        # Set reset high for a millisecond.
        self._rst.value = True
        time.sleep(0.001)
        # Set reset low for 10 milliseconds.
        self._rst.value = False
        time.sleep(0.010)
        # Set reset high again.
        self._rst.value = False

    def display(self):
        """Write display buffer to physical display."""
        self.command(SSD1306_COLUMNADDR)
        self.command(0)              # Column start address. (0 = reset)
        self.command(self.width-1)   # Column end address.
        self.command(SSD1306_PAGEADDR)
        self.command(0)              # Page start address. (0 = reset)
        self.command(self._pages-1)  # Page end address.
        for i in range(0, len(self._buffer), 16):
            control = 0x40   # Co = 0, DC = 0
            self._i2c.write_list(control, self._buffer[i:i+16])

    def image(self, image):
        """Set buffer to value of Python Imaging Library image.  The image should
        be in 1 bit mode and a size equal to the display size.
        """
        if image.mode != '1':
            raise ValueError('Image must be in mode 1.')
        imwidth, imheight = image.size
        if imwidth != self.width or imheight != self.height:
            raise ValueError('Image must be same dimensions as display ({0}x{1}).' \
                .format(self.width, self.height))
        # Grab all the pixels from the image, faster than getpixel.
        pix = image.load()
        # Iterate through the memory pages
        index = 0
        for page in range(self._pages):
            # Iterate through all x axis columns.
            for x in range(self.width):
                # Set the bits for the column of pixels at the current position.
                bits = 0
                # Don't use range here as it's a bit slow
                for bit in [0, 1, 2, 3, 4, 5, 6, 7]:
                    bits = bits << 1
                    bits |= 0 if pix[(x, page*8+7-bit)] == 0 else 1
                # Update buffer byte and increment to next byte.
                self._buffer[index] = bits
                index += 1

    def clear(self):
        """Clear contents of image buffer."""
        self._buffer = [0]*(self.width*self._pages)

    def set_contrast(self, contrast):
        """Sets the contrast of the display.  Contrast should be a value between
        0 and 255."""
        if contrast < 0 or contrast > 255:
            raise ValueError('Contrast must be a value from 0 to 255 (inclusive).')
        self.command(SSD1306_SETCONTRAST)
        self.command(contrast)

    def dim(self, dim):
        """Adjusts contrast to dim the display if dim is True, otherwise sets the
        contrast to normal brightness if dim is False.
        """
        # Assume dim display.
        contrast = 0
        # Adjust contrast based on VCC if not dimming.
        if not dim:
            if self._vccstate == SSD1306_EXTERNALVCC:
                contrast = 0x9F
            else:
                contrast = 0xCF


class SSD1306DeviceDriver(_SSD1306I2CBase):
    def __init__(self, height, width = 128, rst = None, i2c_bus=None, i2c_address=SSD1306_I2C_ADDRESS):
        # Call base class constructor.
        if height == 64:
            super(SSD1306DeviceDriver, self).__init__(128, 64, rst, i2c_bus, i2c_address)
        elif height == 32:
            super(SSD1306DeviceDriver, self).__init__(128, 32, rst, i2c_bus, i2c_address)
        elif height == 16: 
            super(SSD1306DeviceDriver, self).__init__(96, 16, rst, i2c_bus, i2c_address)


    def _initialize(self):
        if self.height == 64:
            self._initialize64()
        elif self.height == 32:
            self._initialize32()
        elif self.height == 16:
            self._initialize16()

    
    def _initialize16(self):
        # 128x32 pixel specific initialization.
        self.command(SSD1306_DISPLAYOFF)                    # 0xAE
        self.command(SSD1306_SETDISPLAYCLOCKDIV)            # 0xD5
        self.command(0x60)                                  # the suggested ratio 0x60
        self.command(SSD1306_SETMULTIPLEX)                  # 0xA8
        self.command(0x0F)
        self.command(SSD1306_SETDISPLAYOFFSET)              # 0xD3
        self.command(0x0)                                   # no offset
        self.command(SSD1306_SETSTARTLINE | 0x0)            # line #0
        self.command(SSD1306_CHARGEPUMP)                    # 0x8D
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x10)
        else:
            self.command(0x14)
        self.command(SSD1306_MEMORYMODE)                    # 0x20
        self.command(0x00)                                  # 0x0 act like ks0108
        self.command(SSD1306_SEGREMAP | 0x1)
        self.command(SSD1306_COMSCANDEC)
        self.command(SSD1306_SETCOMPINS)                    # 0xDA
        self.command(0x02)
        self.command(SSD1306_SETCONTRAST)                   # 0x81
        self.command(0x8F)
        self.command(SSD1306_SETPRECHARGE)                  # 0xd9
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x22)
        else:
            self.command(0xF1)
        self.command(SSD1306_SETVCOMDETECT)                 # 0xDB
        self.command(0x40)
        self.command(SSD1306_DISPLAYALLON_RESUME)           # 0xA4
        self.command(SSD1306_NORMALDISPLAY)     
    
    def _initialize32(self):
        # 128x32 pixel specific initialization.
        self.command(SSD1306_DISPLAYOFF)                    # 0xAE
        self.command(SSD1306_SETDISPLAYCLOCKDIV)            # 0xD5
        self.command(0x80)                                  # the suggested ratio 0x80
        self.command(SSD1306_SETMULTIPLEX)                  # 0xA8
        self.command(0x1F)
        self.command(SSD1306_SETDISPLAYOFFSET)              # 0xD3
        self.command(0x0)                                   # no offset
        self.command(SSD1306_SETSTARTLINE | 0x0)            # line #0
        self.command(SSD1306_CHARGEPUMP)                    # 0x8D
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x10)
        else:
            self.command(0x14)
        self.command(SSD1306_MEMORYMODE)                    # 0x20
        self.command(0x00)                                  # 0x0 act like ks0108
        self.command(SSD1306_SEGREMAP | 0x1)
        self.command(SSD1306_COMSCANDEC)
        self.command(SSD1306_SETCOMPINS)                    # 0xDA
        self.command(0x02)
        self.command(SSD1306_SETCONTRAST)                   # 0x81
        self.command(0x8F)
        self.command(SSD1306_SETPRECHARGE)                  # 0xd9
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x22)
        else:
            self.command(0xF1)
        self.command(SSD1306_SETVCOMDETECT)                 # 0xDB
        self.command(0x40)
        self.command(SSD1306_DISPLAYALLON_RESUME)           # 0xA4
        self.command(SSD1306_NORMALDISPLAY)                 # 0xA6

    
    
    def _initialize64(self):
        # 128x64 pixel specific initialization.
        self.command(SSD1306_DISPLAYOFF)                    # 0xAE
        self.command(SSD1306_SETDISPLAYCLOCKDIV)            # 0xD5
        self.command(0x80)                                  # the suggested ratio 0x80
        self.command(SSD1306_SETMULTIPLEX)                  # 0xA8
        self.command(0x3F)
        self.command(SSD1306_SETDISPLAYOFFSET)              # 0xD3
        self.command(0x0)                                   # no offset
        self.command(SSD1306_SETSTARTLINE | 0x0)            # line #0
        self.command(SSD1306_CHARGEPUMP)                    # 0x8D
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x10)
        else:
            self.command(0x14)
        self.command(SSD1306_MEMORYMODE)                    # 0x20
        self.command(0x00)                                  # 0x0 act like ks0108
        self.command(SSD1306_SEGREMAP | 0x1)
        self.command(SSD1306_COMSCANDEC)
        self.command(SSD1306_SETCOMPINS)                    # 0xDA
        self.command(0x12)
        self.command(SSD1306_SETCONTRAST)                   # 0x81
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x9F)
        else:
            self.command(0xCF)
        self.command(SSD1306_SETPRECHARGE)                  # 0xd9
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x22)
        else:
            self.command(0xF1)
        self.command(SSD1306_SETVCOMDETECT)                 # 0xDB
        self.command(0x40)
        self.command(SSD1306_DISPLAYALLON_RESUME)           # 0xA4
        self.command(SSD1306_NORMALDISPLAY)                 # 0xA6


class SSD1306_128_32_DeviceDriver(_SSD1306I2CBase):
    def __init__(self, rst, dc=None, i2c_bus=None, i2c_address=SSD1306_I2C_ADDRESS):
        # Call base class constructor.
        super(SSD1306_128_32_DeviceDriver, self).__init__(128, 32, rst, i2c_bus, i2c_address)

    def _initialize(self):
        # 128x32 pixel specific initialization.
        self.command(SSD1306_DISPLAYOFF)                    # 0xAE
        self.command(SSD1306_SETDISPLAYCLOCKDIV)            # 0xD5
        self.command(0x80)                                  # the suggested ratio 0x80
        self.command(SSD1306_SETMULTIPLEX)                  # 0xA8
        self.command(0x1F)
        self.command(SSD1306_SETDISPLAYOFFSET)              # 0xD3
        self.command(0x0)                                   # no offset
        self.command(SSD1306_SETSTARTLINE | 0x0)            # line #0
        self.command(SSD1306_CHARGEPUMP)                    # 0x8D
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x10)
        else:
            self.command(0x14)
        self.command(SSD1306_MEMORYMODE)                    # 0x20
        self.command(0x00)                                  # 0x0 act like ks0108
        self.command(SSD1306_SEGREMAP | 0x1)
        self.command(SSD1306_COMSCANDEC)
        self.command(SSD1306_SETCOMPINS)                    # 0xDA
        self.command(0x02)
        self.command(SSD1306_SETCONTRAST)                   # 0x81
        self.command(0x8F)
        self.command(SSD1306_SETPRECHARGE)                  # 0xd9
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x22)
        else:
            self.command(0xF1)
        self.command(SSD1306_SETVCOMDETECT)                 # 0xDB
        self.command(0x40)
        self.command(SSD1306_DISPLAYALLON_RESUME)           # 0xA4
        self.command(SSD1306_NORMALDISPLAY)                 # 0xA6


class SSD1306_96_16_DeviceDriver(_SSD1306I2CBase):
    def __init__(self, rst=None, i2c_bus=None, i2c_address=SSD1306_I2C_ADDRESS):
        # Call base class constructor.
        super(SSD1306_96_16_DeviceDriver, self).__init__(96, 16, rst, i2c_bus, i2c_address)

    def _initialize(self):
        # 128x32 pixel specific initialization.
        self.command(SSD1306_DISPLAYOFF)                    # 0xAE
        self.command(SSD1306_SETDISPLAYCLOCKDIV)            # 0xD5
        self.command(0x60)                                  # the suggested ratio 0x60
        self.command(SSD1306_SETMULTIPLEX)                  # 0xA8
        self.command(0x0F)
        self.command(SSD1306_SETDISPLAYOFFSET)              # 0xD3
        self.command(0x0)                                   # no offset
        self.command(SSD1306_SETSTARTLINE | 0x0)            # line #0
        self.command(SSD1306_CHARGEPUMP)                    # 0x8D
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x10)
        else:
            self.command(0x14)
        self.command(SSD1306_MEMORYMODE)                    # 0x20
        self.command(0x00)                                  # 0x0 act like ks0108
        self.command(SSD1306_SEGREMAP | 0x1)
        self.command(SSD1306_COMSCANDEC)
        self.command(SSD1306_SETCOMPINS)                    # 0xDA
        self.command(0x02)
        self.command(SSD1306_SETCONTRAST)                   # 0x81
        self.command(0x8F)
        self.command(SSD1306_SETPRECHARGE)                  # 0xd9
        if self._vccstate == SSD1306_EXTERNALVCC:
            self.command(0x22)
        else:
            self.command(0xF1)
        self.command(SSD1306_SETVCOMDETECT)                 # 0xDB
        self.command(0x40)
        self.command(SSD1306_DISPLAYALLON_RESUME)           # 0xA4
        self.command(SSD1306_NORMALDISPLAY)                 # 0xA6